/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright (c) 1990, 2010, Oracle and/or its affiliates. All rights reserved.
 * Copyright 2012 Nexenta Systems, Inc.  All rights reserved.
 */

/*
 * For SUNWnskit - version 1.1
 */

#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include <pwd.h>
#include <rpcsvc/ypclnt.h>
#include "util.h"
#include "table.h"
#include "getgroup.h"
#include "revnetgroup.h"

#define	MAXDOMAINLEN 256
#define	MAXGROUPLEN  131072
#define	MAXKEYLEN    512

/*
 * Reverse the netgroup file. A flag of "-u" means reverse by username,
 * one of "-h" means reverse by hostname. Each line in the output file
 * will begin with a key formed by concatenating the host or user name
 * with the domain name. The key will be followed by a tab, then the
 * comma-separated, newline-terminated list of groups to which the
 * user or host belongs.
 *
 * Exception: Groups to which everyone belongs (universal groups) will
 * not be included in the list.  The universal groups will be listed under
 * the special name "*".
 *
 * Thus to find out all the groups that user "foo" of domain "bar" is in,
 * lookup the groups under  foo.bar, foo.*, *.bar and *.*.
 *
 */

static char *nextgroup(revhandle_t *, tablelist *);
static void storegroup(char *group, revhandle_t *hdl);
static void enter(char *name, char *group, revhandle_t *hdl);
static void appendgroup(groupentrylist grlist, char *group);
static groupentrylist newentry(char *name, char *group);
static void loadtable(FILE *nf, revhandle_t *hdl);
static void dumptable(FILE *out, revhandle_t *hdl);
static void free_table(revhandle_t *hdl);

int
revnetgroup_handle(FILE *fin, FILE *fout, boolean_t byuser)
{
	char *group;
	revhandle_t hdl = {0};
	tablelist walker = NULL;

	hdl.rh_byuser = byuser;
	loadtable(fin, &hdl);

	while (group = nextgroup(&hdl, &walker))
		storegroup(group, &hdl);

	dumptable(fout, &hdl);
	free_table(&hdl);
	return (0);
}

/*
 *	Get the next netgroup from /etc/netgroup
 */
static char *
nextgroup(revhandle_t *hdl, tablelist *next)
{
	tablelist entry = *next;
	uint_t i;

	if (entry == NULL) {
		for (i = hdl->rh_index; i < TABLESIZE; i++) {
			entry = hdl->ngtable[i];
			if (entry != NULL) {
				hdl->rh_index = i + 1;
				break;
			}
		}
	}

	if (entry == NULL) {
		/* out of table */
		hdl->rh_index = 0;
		return (NULL);
	}

	*next = entry->next;
	return (entry->key);
}


static void
free_tablelist(tablelist e)
{
	tablelist next;

	do {
		next = e->next;

		free(e->key);
		free(e->datum);
		free(e);
	} while ((e = next) != NULL);
}

static void
free_stringnode(stringnode *s)
{
	stringnode *next;

	do {
		next = s->s_next;

		free(s);
	} while ((s = next) != NULL);
}

static void
free_groupentrylist(groupentrylist e)
{
	groupentrylist next;

	do {
		next = e->next;

		free_stringnode(e->groups);
		free(e->name);
		free(e);
	} while ((e = next) != NULL);
}

static void
free_table(revhandle_t *hdl)
{
	int i;

	/* ngtable */
	for (i = 0; i < TABLESIZE; i++) {
		tablelist e;

		if ((e = hdl->ngtable[i]) == NULL)
			continue;

		free_tablelist(e);
		hdl->ngtable[i] = NULL;
	}

	/* grouptable */
	for (i = 0; i < TABLESIZE; i++) {
		groupentrylist e;

		if ((e = hdl->grouptable[i]) == NULL)
			continue;

		free_groupentrylist(e);
		hdl->grouptable[i] = NULL;
	}
}

/*
 * Dump out all of the stored info into a file
 */
static void
dumptable(FILE *out, revhandle_t *hdl)
{
	int i;
	groupentrylist entry;
	stringnode *groups;

	for (i = 0; i < TABLESIZE; i++) {
		if (entry = hdl->grouptable[i]) {
			while (entry) {
				fputs(entry->name, out);
				putc('\t', out);
				for (groups = entry->groups; groups;
				    groups = groups->s_next) {
					fputs(groups->str, out);
					if (groups->s_next) {
						putc(',', out);
					}
				}
				putc('\n', out);
				entry = entry->next;
			}
		}
	}
}



/*
 *	Add a netgroup to a user's list of netgroups
 */
static void
storegroup(char *group, revhandle_t *hdl)
{
	char key[MAXKEYLEN];
	struct grouplist *glist;

	doit(group, (struct list *)NULL, hdl);

	for (glist = hdl->grouplist; glist; glist = glist->gl_nxt) {
		const char *name;	/* username or hostname */
		const char *domain;

		name = hdl->rh_byuser ? glist->gl_name : glist->gl_machine;
		if (!name) {
			name = "*";
		} else if (!isalnum(*name) && *name != '_') {
			continue;
		}
		domain = glist->gl_domain;
		if (!domain) {
			domain = "*";
		}
		(void) snprintf(key, sizeof (key), "%s.%s", name, domain);
		enter(key, group, hdl);
	}
	freegrouplist(hdl);
}



static groupentrylist
newentry(char *name, char *group)
{
	groupentrylist new;

	new = MALLOC(groupentrynode);

	STRCPY(new->name, name);

	new->groups = MALLOC(stringnode);
	new->groups->str = group;
	new->groups->s_next = NULL;

	new->next = NULL;
	return (new);
}

static void
appendgroup(groupentrylist grlist, char *group)
{
	stringnode *cur, *prev;

	for (cur = grlist->groups; cur; prev = cur, cur = cur->s_next) {
		if (strcmp(group, cur->str) == 0) {
			return;
		}
	}
	prev->s_next = MALLOC(stringnode);
	cur = prev->s_next;
	cur->str = group;
	cur->s_next = NULL;
}

static void
enter(char *name, char *group, revhandle_t *hdl)
{
	int key;
	groupentrylist gel;
	groupentrylist gelprev;

	key = tablekey(name);
	if (hdl->grouptable[key] == NULL) {
		hdl->grouptable[key] = newentry(name, group);
	} else {
		gel = hdl->grouptable[key];
		while (gel && strcmp(gel->name, name)) {
			gelprev = gel;
			gel = gel->next;
		}
		if (gel) {
			appendgroup(gel, group);
		} else {
			gelprev->next = newentry(name, group);
		}
	}
}

/*
 * Load up a hash table with the info in /etc/netgroup
 */
static void
loadtable(FILE *nf, revhandle_t *hdl)
{
	char *buf;
	char *p;
	char *group;
	char *line;

	buf = malloc(MAXGROUPLEN);
	if (buf == NULL)
		return;

	while (getaline(buf, MAXGROUPLEN, nf)) {
		/* skip leading blanks */
		for (p = buf; *p && isspace((int)*p); p++)
			;
		for (; *p && *p != '#' && *p != ' ' && *p != '\t'; p++)
			;
		if (*p == EOS || *p == '#')
			continue;
		*p++ = EOS;

		while (*p == ' ' || *p == '\t') {
			p++;
		}
		if (*p == EOS || *p == '#')
			continue;

		STRCPY(group, buf);
		STRCPY(line, p);
		store(hdl->ngtable, group, line);
	}

	free(buf);
}
